@@ -22,8 +22,8 @@
  * is now "desktop widget".
  * 
  * As with other custom GTK+ widgets, drawing happens only in a handler for the 
- * expose-event signal. This is triggered by a timeout handler that invalidates
- * the desktop widget's area.
+ * expose-event signal. This is triggered by invalidating the desktop widget's
+ * area.
  *
  */
 
@@ -31,21 +31,22 @@
 // TODO:
 // - sprinkle with type-checking macros and g_assert
 // - save settings in gconf
-// - handle button_press, button_release, leave_notify
-//   - highlight on button press, perform action on release
+// - use HildonCaption in settings dialog
+// - set .so version
 
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
 
+
 #include <stdlib.h>
 #include <errno.h>
 #include <unistd.h>
 #include <string.h>
 
 #include <glib.h>
-#include <glib/gi18n-lib.h> /* Declares _(). */
+#include <glib/gi18n-lib.h> /* Declares _() - alias for gettext() */
 
 #include <gtk/gtk.h>
 
@@ -59,7 +60,7 @@
 #include "ccchh-dooris-widget.h"
 
 
-// #define DEBUG
+//#define DEBUG
 
 #ifdef DEBUG
 
@@ -72,7 +73,7 @@
 #else
 
 // check g_debug statements for syntax errors but don't emit code
-// (compiler should optimize it)
+// (compiler should optimize it away)
 #define g_debug(fmt...) do { g_message(fmt); } while (0)
 
 #define USER_AGENT_VERSION_STRING (some version number)
@@ -80,6 +81,8 @@
 #endif // DEBUG
 
 
+
+
 /** This is like G_DEFINE_TYPE(), but it also 
  * implements standard functions such as hd_plugin_module_load(), 
  * which hildon-desktop expects to be able to call on this library.
@@ -88,9 +91,8 @@ HD_DEFINE_PLUGIN_MODULE (CCCHHDoorIsDesktopWidget, ccchh_dooris_desktop_widget,
   HD_TYPE_HOME_PLUGIN_ITEM)
 
 
-// private functions, only called from within the widget
 
- 
+
 // Code to be reworked:
 
 /*
@@ -117,6 +119,9 @@ static void gtk_sleep (unsigned int ms)
 */
 
 
+
+
+// lock fetch mutex, redraw transparent button
 static inline gboolean
 fetch_update_lock (CCCHHDoorIsDesktopWidget *self)
 {
@@ -134,6 +139,7 @@ fetch_update_lock (CCCHHDoorIsDesktopWidget *self)
 }
 
 
+// unlock fetch mutex, redraw opaque button
 static inline void
 fetch_update_unlock (CCCHHDoorIsDesktopWidget *self)
 {
@@ -144,8 +150,7 @@ fetch_update_unlock (CCCHHDoorIsDesktopWidget *self)
 }
 
 
-// initialize async i/o
-// callback: fetch_update_async_open_cb()
+// initialize async update from server, lock mutex
 static void
 fetch_update_start (CCCHHDoorIsDesktopWidget *self)
 {
@@ -163,13 +168,6 @@ fetch_update_start (CCCHHDoorIsDesktopWidget *self)
 
 	g_debug("ccchh-dooris-widget: begin update from URL: %s", self->url);
 
-
-/*
-	if (self->gvfs_handle) {
-		g_object_unref(self->gvfs_handle);
-		self->gvfs_handle = NULL;
-	}
-*/
 	// TODO: set timeout to cancel async open
 
 	gnome_vfs_async_open(&self->gvfs_handle, self->url, GNOME_VFS_OPEN_READ,
@@ -177,6 +175,7 @@ fetch_update_start (CCCHHDoorIsDesktopWidget *self)
 }
 
 
+// handle async open callback, begin read
 static void
 fetch_update_async_open_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
     gpointer data)
@@ -187,20 +186,25 @@ fetch_update_async_open_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 
 	if (result != GNOME_VFS_OK) {
 		g_warning("ccchh-dooris-widget: error opening URL: %s", gnome_vfs_result_to_string (result));
-		fetch_update_unlock(self);
+		fetch_update_process(self, DOORIS_FNORD, 0L, 0L);
 		return;
 	}
 
+/*
 	if (self->gvfs_input_buffer) {
 		g_free(self->gvfs_input_buffer);
 		self->gvfs_input_buffer = NULL;
 	}
+*/
 
-	if (!(self->gvfs_input_buffer = g_try_malloc0(MAX_HTTP_SIZE+1))) {
+	// reuse input buffer, ensure it has at least MAX_HTTP_SIZE+1
+	if (!(self->gvfs_input_buffer = g_try_realloc(self->gvfs_input_buffer, MAX_HTTP_SIZE+1))) {
 		g_warning("ccchh-dooris-widget: failed to allocate input buffer");
-		fetch_update_unlock(self);
+		fetch_update_process(self, DOORIS_FNORD, 0L, 0L);
 		return;
 	}
+	// initialize input buffer with zeroes, so strtok won't read beyond its end
+	memset(self->gvfs_input_buffer, 0, MAX_HTTP_SIZE+1);
 
 	// TODO: set timeout to cancel async read
 
@@ -209,6 +213,7 @@ fetch_update_async_open_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 }
 
 
+// handle async read callback, begin close, parse input
 static void
 fetch_update_async_read_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 	gpointer buffer, GnomeVFSFileSize bytes_requested, GnomeVFSFileSize bytes_read,
@@ -220,7 +225,7 @@ fetch_update_async_read_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 
 	if (result != GNOME_VFS_OK) {
 		g_warning("ccchh-dooris-widget: read error: %s", gnome_vfs_result_to_string (result));
-		fetch_update_unlock(self);
+		fetch_update_process(self, DOORIS_FNORD, 0L, 0L);
 		return;
 	}
 
@@ -246,7 +251,7 @@ fetch_update_async_read_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 
 	if (!(line = strtok_r(self->gvfs_input_buffer, "\n", &tokctx))) {
 		g_warning("ccchh-dooris-widget: first line of response not found");
-		process_update(self, new_state, new_change_ts, new_report_ts);
+		fetch_update_process(self, new_state, new_change_ts, new_report_ts);
 		return;
 	}
 
@@ -258,7 +263,7 @@ fetch_update_async_read_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 		g_debug("ccchh-dooris-widget: current state: open");
 	} else {
 		g_warning("ccchh-dooris-widget: could not parse response");
-		process_update(self, new_state, new_change_ts, new_report_ts);
+		fetch_update_process(self, new_state, new_change_ts, new_report_ts);
 		return;
 	}
 
@@ -288,19 +293,22 @@ fetch_update_async_read_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 		}
 	}
 
+/*
 	// TODO: reuse small buffers
 	g_free(self->gvfs_input_buffer);
 	self->gvfs_input_buffer = NULL;
+*/
 
-	process_update(self, new_state, new_change_ts, new_report_ts);
+	fetch_update_process(self, new_state, new_change_ts, new_report_ts);
 }
 
 
+// handle async close callback
 static void
 fetch_update_async_close_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 	gpointer data)
 {
-	CCCHHDoorIsDesktopWidget *self = CCCHH_DOORIS_DESKTOP_WIDGET (data);
+//	CCCHHDoorIsDesktopWidget *self = CCCHH_DOORIS_DESKTOP_WIDGET (data);
 
 	g_debug("ccchh-dooris-widget: fetch_update_async_close_cb()");
 
@@ -308,9 +316,75 @@ fetch_update_async_close_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 		g_debug("ccchh-dooris-widget: error on close: %s", gnome_vfs_result_to_string (result));
 		// TODO: what now?
 	}
+}
 
-// TODO: think again
-//	g_usleep(100000);
+
+// update internal state, unlock mutex
+static void
+fetch_update_process (CCCHHDoorIsDesktopWidget *self, dooris_state_t new_state,
+	gulong new_change_ts, gulong new_report_ts)
+{
+	g_debug("ccchh-dooris-widget: fetch_update_process()");
+
+	self->current_state = new_state;
+	self->last_change_ts = new_change_ts;
+	self->last_report_ts = new_report_ts;
+
+	if (new_state == self->last_state) {
+		g_debug("ccchh-dooris-widget: state unchanged");
+
+		if (self->force_banner) {
+			self->force_banner = FALSE;
+			show_status_banner(self);
+		}
+
+		fetch_update_unlock(self);
+		return;
+	}
+
+	// release old icon surface
+	if (self->image_current) {
+		cairo_surface_destroy(self->image_current);
+		self->image_current = NULL;
+	}
+
+	// set new icon surface
+	switch (self->current_state) {
+		case DOORIS_OPEN:
+			self->image_current = cairo_image_surface_create_from_png(IMAGE_DOORIS_OPEN);
+			g_debug("ccchh-dooris-widget: image_current set to open");
+			break;
+		case DOORIS_CLOSED:
+			self->image_current = cairo_image_surface_create_from_png(IMAGE_DOORIS_CLOSED);
+			g_debug("ccchh-dooris-widget: image_current set to closed");
+			break;
+		default:
+			self->image_current = cairo_image_surface_create_from_png(IMAGE_DOORIS_FNORD);
+			g_debug("ccchh-dooris-widget: image_current set to fnord");
+			break;
+	}
+
+	self->last_state = self->current_state;
+
+	// state is valid and has changed since last known state
+	if (self->current_state != DOORIS_FNORD && self->current_state != self->last_known_state) {
+
+		self->last_known_state = self->current_state;
+		g_debug("ccchh-dooris-widget: state changed");
+
+		// state changed: redraw icon
+		force_redraw(self);
+
+		// flash banner if notifications are enabled
+		if (self->notify) {
+			self->force_banner = TRUE;
+		}
+	}
+
+	if (self->force_banner) {
+		self->force_banner = FALSE;
+		show_status_banner(self);
+	}
 
 	fetch_update_unlock(self);
 }
@@ -401,84 +475,6 @@ remove_connection_handler(CCCHHDoorIsDesktopWidget *self)
 
 
 
-// update internal state
-
-static void
-process_update (CCCHHDoorIsDesktopWidget *self, dooris_state_t new_state,
-	gulong new_change_ts, gulong new_report_ts)
-{
-	static GStaticMutex process_lock = G_STATIC_MUTEX_INIT;
-
-	g_debug("ccchh-dooris-widget: process_update()");
-
-	if (!g_static_mutex_trylock(&process_lock)) {
-		g_warning("ccchh-dooris-widget: BUG: process_update() lock set, not updating");
-		return;
-	}
-
-	self->current_state = new_state;
-	self->last_change_ts = new_change_ts;
-	self->last_report_ts = new_report_ts;
-
-	if (new_state == self->last_state) {
-		g_debug("ccchh-dooris-widget: state unchanged");
-
-		if (self->force_banner) {
-			self->force_banner = FALSE;
-			show_status_banner(self);
-		}
-
-		g_static_mutex_unlock(&process_lock);
-		return;
-	}
-
-	// release old icon surface
-	if (self->image_current) {
-		cairo_surface_destroy(self->image_current);
-		self->image_current = NULL;
-	}
-
-	// set new icon surface
-	switch (self->current_state) {
-		case DOORIS_OPEN:
-			self->image_current = cairo_image_surface_create_from_png(IMAGE_DOORIS_OPEN);
-			g_debug("ccchh-dooris-widget: image_current set to open");
-			break;
-		case DOORIS_CLOSED:
-			self->image_current = cairo_image_surface_create_from_png(IMAGE_DOORIS_CLOSED);
-			g_debug("ccchh-dooris-widget: image_current set to closed");
-			break;
-		default:
-			self->image_current = cairo_image_surface_create_from_png(IMAGE_DOORIS_FNORD);
-			g_debug("ccchh-dooris-widget: image_current set to fnord");
-			break;
-	}
-
-	self->last_state = self->current_state;
-
-	// state is valid and has changed since last known state
-	if (self->current_state != DOORIS_FNORD && self->current_state != self->last_known_state) {
-
-		self->last_known_state = self->current_state;
-		g_debug("ccchh-dooris-widget: state changed");
-
-		// state changed: redraw icon
-		force_redraw(self);
-
-		// flash banner if notifications are enabled
-		if (self->notify) {
-			self->force_banner = TRUE;
-		}
-	}
-
-	if (self->force_banner) {
-		self->force_banner = FALSE;
-		show_status_banner(self);
-	}
-
-	g_static_mutex_unlock(&process_lock);
-}
-
 
 static void
 draw (CCCHHDoorIsDesktopWidget *self, cairo_t *cr)
@@ -507,6 +503,13 @@ draw (CCCHHDoorIsDesktopWidget *self, cairo_t *cr)
 		cairo_surface_destroy(image_overlay);
 	}
 
+	//
+	if (self->highlighted) {
+		cairo_set_source_rgba(cr, 1.0, 1.0, 1.0, 0.6);
+		cairo_mask_surface(cr, self->image_current, 5, 5);
+//		cairo_paint_with_alpha(cr, self->current_alpha);
+	}
+
 	cairo_restore(cr);
 }
 
@@ -645,8 +648,7 @@ ccchh_dooris_desktop_widget_expose_event (GtkWidget *widget, GdkEventExpose *eve
 }
 
 
-// handle the button_press event
-// TODO: highlight on button_press, call update() on button_release
+// handle the button press event, highlight icon
 static gboolean 
 ccchh_dooris_desktop_widget_on_button_press_event (GtkWidget *widget, 
   GdkEventButton * event G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
@@ -655,12 +657,45 @@ ccchh_dooris_desktop_widget_on_button_press_event (GtkWidget *widget,
 
 	g_debug("ccchh-dooris-widget: ccchh_dooris_desktop_widget_on_button_press_event()");
  
-	// update() must be called only once at a time
+	// TODO: do we really need a button lock? fetch_update locks itself now
 	if (!g_static_mutex_trylock(&(self->button_lock))) {
 		g_debug("ccchh-dooris-widget: button_lock set, not updating");
 		return TRUE;
 	}
 
+	self->highlighted = TRUE;
+	force_redraw(self);
+
+	return TRUE; /* Prevent further handling of this signal. */
+}
+
+// handle the leave notify event, remove highlight
+static gboolean 
+ccchh_dooris_desktop_widget_on_button_leave_event (GtkWidget *widget, 
+  GdkEventButton * event G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
+{
+	CCCHHDoorIsDesktopWidget *self = CCCHH_DOORIS_DESKTOP_WIDGET (widget);
+
+	g_debug("ccchh-dooris-widget: ccchh_dooris_desktop_widget_on_button_leave_event()");
+ 
+	self->highlighted = FALSE;
+	force_redraw(self);
+
+	g_static_mutex_unlock(&(self->button_lock));
+
+	return TRUE; /* Prevent further handling of this signal. */
+}
+
+
+// handle the button release event, begin update, remove highlight
+static gboolean 
+ccchh_dooris_desktop_widget_on_button_release_event (GtkWidget *widget, 
+  GdkEventButton * event G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
+{
+	CCCHHDoorIsDesktopWidget *self = CCCHH_DOORIS_DESKTOP_WIDGET (widget);
+
+	g_debug("ccchh-dooris-widget: ccchh_dooris_desktop_widget_on_button_release_event()");
+ 
 	// request connection from libconic, if offline and
 	// force_connection is enabled
 	// TODO: this does not work immediately, the call to update()
@@ -682,6 +717,9 @@ ccchh_dooris_desktop_widget_on_button_press_event (GtkWidget *widget,
 		start_update_timer(self);
 	}
 
+	self->highlighted = FALSE;
+	force_redraw(self);
+
 	g_static_mutex_unlock(&(self->button_lock));
 
 	return TRUE; /* Prevent further handling of this signal. */
@@ -1054,7 +1092,6 @@ ccchh_dooris_desktop_widget_finalize (GObject *object)
 	// other parts of the desktop might still use it
  
 	if (self->url) {
-//		g_free(self->url);
 		free(self->url);
 		self->url = NULL;
 	}
@@ -1162,13 +1199,20 @@ ccchh_dooris_desktop_widget_init (CCCHHDoorIsDesktopWidget *self)
 	/* Allow this widget to handle button-press events:
 	 * Note that gtk_widget_add_events would only work after the widget is realized:
 	 */
-	gint mask = gtk_widget_get_events (GTK_WIDGET (self)) | GDK_BUTTON_PRESS_MASK;
+	gint mask = gtk_widget_get_events (GTK_WIDGET (self)) | GDK_BUTTON_PRESS_MASK |
+		GDK_BUTTON_RELEASE_MASK | GDK_LEAVE_NOTIFY_MASK;
 	gtk_widget_set_events (GTK_WIDGET (self), mask);
      
 	/* Handle a mouse button press: */
 	g_signal_connect (self, "button-press-event", 
 		G_CALLBACK (ccchh_dooris_desktop_widget_on_button_press_event), NULL);
 
+	g_signal_connect (self, "button-release-event", 
+		G_CALLBACK (ccchh_dooris_desktop_widget_on_button_release_event), NULL);
+
+	g_signal_connect (self, "leave-notify-event", 
+		G_CALLBACK (ccchh_dooris_desktop_widget_on_button_leave_event), NULL);
+
 	/* Specify that a settings button should be shown in layout mode, 
 	 * and handle a request to configure the settings:
 	 */
