@@ -66,6 +66,7 @@ typedef struct {
 	cairo_surface_t *image_current;
 	gboolean force_banner;
 	gboolean connected;
+	gboolean highlighted;
 	gchar *disconnecting;
 	gdouble current_alpha;
 	GStaticMutex button_lock;
@@ -132,8 +133,8 @@ fetch_update_async_close_cb (GnomeVFSAsyncHandle *handle, GnomeVFSResult result,
 
 
 static void
-process_update (CCCHHDoorIsDesktopWidget *self, dooris_state_t new_state,
-	gulong new_change_ts, gulong new_report_ts);
+fetch_update_process (CCCHHDoorIsDesktopWidget *self, dooris_state_t new_state,
+    gulong new_change_ts, gulong new_report_ts);
 
 
 static void
